/*
 * Zorin Panel: Various settings and changes for the panel in Zorin Desktop.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Credits:
 * This file is based on code from the Dash to Dock extension by micheleg
 * and code from the Dash to Panel extension by Jason DeRose (as at
 * commit 6e53889082eef4eed9cdc1c496e90a6f8450d1fd)
 * 
 * Code to re-anchor the panel was taken from Thoma5 BottomPanel:
 * https://github.com/Thoma5/gnome-shell-extension-bottompanel
 * 
 * Pattern for moving clock based on Frippery Move Clock by R M Yorston
 * http://frippery.org/extensions/
 * 
 * Some code was also adapted from the upstream Gnome Shell source code.
 */

const Me = imports.misc.extensionUtils.getCurrentExtension();
const Clutter = imports.gi.Clutter;
const Utils = Me.imports.utils;
const Lang = imports.lang;
const Gio = imports.gi.Gio;
const Main = imports.ui.main;
const Mainloop = imports.mainloop;
const Panel = imports.ui.panel;
const St = imports.gi.St;
const Meta = imports.gi.Meta;
const DND = imports.ui.dnd;
const Shell = imports.gi.Shell;
const ViewSelector = imports.ui.viewSelector;
const DateMenu = imports.ui.dateMenu;
const PanelMenu = imports.ui.panelMenu;

const Intellihide = Me.imports.intellihide;
const RemoveArrows = Me.imports.removeArrows;

var DYNAMIC_TRANSPARENCY_DISTANCE = 5;
var SHOW_DESKTOP_ICON = Me.path + '/show-desktop-symbolic.svg';

let tracker = Shell.WindowTracker.get_default();

var zorinPanelWrapper = new Lang.Class({
    Name: 'ZorinPanel.PanelWrapper',

    _init: function(panelManager, monitor, panel, panelBox, isSecondary) {
        this.panelManager = panelManager;
        this._zorinPanelSettings = panelManager._zorinPanelSettings;

        this.monitor = monitor;
        this.panel = panel;
        this.panelBox = panelBox;
        this.isSecondary = isSecondary;
    },

    enable : function() {   
        this._oldPanelHeight = this.panel.actor.get_height();

        // The overview uses the this.panel height as a margin by way of a "ghost" transparent Clone
        // This pushes everything down, which isn't desired when the this.panel is moved to the bottom
        // I'm adding a 2nd ghost this.panel and will resize the top or bottom ghost depending on the this.panel position
        this._myPanelGhost = new St.Bin({ 
            child: new Clutter.Clone({ source: this.panel.actor }),
            reactive: false,
            opacity: 0 
        });
        Main.overview._overview.add_actor(this._myPanelGhost);

        this._removeArrows = new RemoveArrows.zorinPanelRemoveArrows(this.panel);
        this._removeArrows.enable();

        this._setPanelPosition();
        
        this._HeightNotifyListener = this.panelBox.connect("notify::height", Lang.bind(this, function(){
            this._setPanelPosition();
        }));

        // this is to catch changes to the window scale factor
        this._ScaleFactorListener = St.ThemeContext.get_for_stage(global.stage).connect("changed", Lang.bind(this, function () { 
            this._setPanelPosition();
        }));
        
        this._setActivitiesButtonVisible(this._zorinPanelSettings.get_boolean('show-activities-button'));
        this._setClockLocation(this._zorinPanelSettings.get_string('location-clock'));
        this._displayShowDesktopButton(this._zorinPanelSettings.get_boolean('show-showdesktop-button'));

        // Since Gnome 3.8 dragging an app without having opened the overview before cause the attemp to
        //animate a null target since some variables are not initialized when the viewSelector is created
        if(Main.overview.viewSelector._activePage == null)
            Main.overview.viewSelector._activePage = Main.overview.viewSelector._workspacesPage;

        this.startIntellihideId = Mainloop.timeout_add(2000, () => {
            this.startIntellihideId = 0;
            this.intellihide = new Intellihide.Intellihide(this);
        });

        this._signalsHandler = new Utils.GlobalSignalsHandler();
        this._signalsHandler.add(
            [
                this.panel._rightBox,
                'actor-added',
                Lang.bind(this, function() {
                    this._setClockLocation(this._zorinPanelSettings.get_string('location-clock'));
                })
            ],
            [
                this.panel._centerBox,
                'actor-added',
                () => this._setClockLocation(this._zorinPanelSettings.get_string('location-clock'))
            ]
        );
        
        this.panel.handleDragOver = Lang.bind(this.panel, function(source, actor, x, y, time) {
            if (source == Main.xdndHandler) {
                
                // open overview so they can choose a window for focusing
                // and ultimately dropping dragged item onto
                if(Main.overview.shouldToggleByCornerOrButton())
                    Main.overview.show();
            }
            
            return DND.DragMotionResult.CONTINUE;
        });

        this.panel._zorinPanelSettings = this._zorinPanelSettings;
        this.panel.oldUpdateSolidStyle = this.panel._updateSolidStyle;
        this.panel._updateSolidStyle = Lang.bind(this.panel, function() {
            if (this._zorinPanelSettings.get_boolean('dynamic-transparency')) {
                if (this.actor.has_style_pseudo_class('overview') || !Main.sessionMode.hasWindows) {
                    this._removeStyleClassName('solid');
                    return;
                }

                if (!Main.layoutManager.primaryMonitor)
                    return;

                /* Get all the windows in the active workspace that are in the primary monitor and visible */
                let workspaceManager = global.workspace_manager;
                let activeWorkspace = workspaceManager.get_active_workspace();
                let windows = activeWorkspace.list_windows().filter(metaWindow => {
                    return metaWindow.is_on_primary_monitor() &&
                           metaWindow.showing_on_its_workspace() &&
                           metaWindow.get_window_type() != Meta.WindowType.DESKTOP;
                });

                /* Check if at least one window is near enough to the panel */
                let [, panelTop] = this.actor.get_transformed_position();
                let panelBottom = panelTop + this.actor.get_height();
                let scale = St.ThemeContext.get_for_stage(global.stage).scale_factor;
                let isNearEnough = windows.some(metaWindow => {
                    if (this.actor.has_style_class_name('bottom')) {
                        let verticalPosition = metaWindow.get_frame_rect().y + metaWindow.get_frame_rect().height;
                        return verticalPosition > panelTop - DYNAMIC_TRANSPARENCY_DISTANCE * scale;
                    } else {
                        let verticalPosition = metaWindow.get_frame_rect().y;
                        return verticalPosition < panelBottom + DYNAMIC_TRANSPARENCY_DISTANCE * scale;
                    }
                });

                if (isNearEnough)
                    this._addStyleClassName('solid');
                else
                    this._removeStyleClassName('solid');
            } else {
                this._addStyleClassName('solid');
            }

        });

	// Since we are usually visible but not usually changing, make sure
	// most repaint requests don't actually require us to repaint anything.
	// This saves significant CPU when repainting the screen.
        this.panel.actor.set_offscreen_redirect(Clutter.OffscreenRedirect.ALWAYS);

        // Set initial panel style
        this.panel._updateSolidStyle();

        this._bindSettingsChanges();
    },

    disable: function () {
        this._removePanelGhost();

        this._signalsHandler.destroy();

        if (this.startIntellihideId) {
            Mainloop.source_remove(this.startIntellihideId);
            this.startIntellihideId = 0;
        } else {
            this.intellihide.destroy();
        }

        // remove this.panel styling
        if(this._HeightNotifyListener !== null) {
            this.panelBox.disconnect(this._HeightNotifyListener);
        }
        if(this._ScaleFactorListener !== null) {
            St.ThemeContext.get_for_stage(global.stage).disconnect(this._ScaleFactorListener);
        }

        for (let i = 0; i < this._zorinPanelSettingsSignalIds.length; ++i) {
            this._zorinPanelSettings.disconnect(this._zorinPanelSettingsSignalIds[i]);
        }

        this._removeTopLimit();

        this.panel._updateSolidStyle = this.panel.oldUpdateSolidStyle;
        this.panel.oldUpdateSolidStyle = null;
        this.panel._zorinPanelSettings = null;

        this.panel.actor.set_height(this._oldPanelHeight);

        if(this.panel.actor.has_style_class_name('top')) {
            this.panel.actor.remove_style_class_name('top');
        } else if(this.panel.actor.has_style_class_name('bottom')) {
            this.panel.actor.remove_style_class_name('bottom');
        }
            
        Main.overview._panelGhost.set_height(this._oldPanelHeight);
        this._setActivitiesButtonVisible(true);
        this._setClockLocation("BUTTONSLEFT");
        this._displayShowDesktopButton(false);

        this._removeArrows.disable();

        this.panel = null;
        this._signalsHandler = null;
        this._HeightNotifyListener = null;
    },

    _bindSettingsChanges: function() {
        this._zorinPanelSettingsSignalIds = [
            this._zorinPanelSettings.connect('changed::panel-size', Lang.bind(this, function() {
                this._setPanelPosition();
            })),

            this._zorinPanelSettings.connect('changed::show-activities-button', Lang.bind(this, function() {
                this._setActivitiesButtonVisible(this._zorinPanelSettings.get_boolean('show-activities-button'));
            })),

            this._zorinPanelSettings.connect('changed::location-clock', Lang.bind(this, function() {
                this._setClockLocation(this._zorinPanelSettings.get_string('location-clock'));
            })),

            this._zorinPanelSettings.connect('changed::show-showdesktop-button', Lang.bind(this, function() {
                this._displayShowDesktopButton(this._zorinPanelSettings.get_boolean('show-showdesktop-button'));
            })),

            this._zorinPanelSettings.connect('changed::dynamic-transparency', Lang.bind(this, function() {
                this.panel._updateSolidStyle();
            }))
        ];
    },
    
    _removePanelGhost: function() {
        if (this._myPanelGhost.get_parent()) {
            Main.overview._overview.remove_actor(this._myPanelGhost);
        }
    },

    _setPanelPosition: function() {
        let scaleFactor = St.ThemeContext.get_for_stage(global.stage).scale_factor;
        let size = this._zorinPanelSettings.get_int('panel-size');
        
        if(scaleFactor)
            size = size*scaleFactor;

        this.panel.actor.set_height(size);

        let position = this._zorinPanelSettings.get_string('panel-position');
        let isTop = position == "TOP";

        Main.overview._panelGhost.set_height(isTop ? size : 0);
        this._myPanelGhost.set_height(isTop ? 0 : size);

        if(isTop) {
            this.panelBox.set_position(this.monitor.x, this.monitor.y);

            this._removeTopLimit();
            
            // styles for theming
            if(this.panel.actor.has_style_class_name('bottom'))
                this.panel.actor.remove_style_class_name('bottom');

            if(!this.panel.actor.has_style_class_name('top'))
                this.panel.actor.add_style_class_name('top');
        } else {
            this.panelBox.set_position(this.monitor.x, this.monitor.y + this.monitor.height - this.panelBox.height);

            if (!this._topLimit) {
                this._topLimit = new St.BoxLayout({ name: 'topLimit', vertical: true });
                Main.layoutManager.addChrome(this._topLimit, { affectsStruts: true, trackFullscreen: true });
            }

            this._topLimit.set_position(this.monitor.x, this.monitor.y);
            this._topLimit.set_size(this.monitor.width, -1);

            // styles for theming
            if(this.panel.actor.has_style_class_name('top'))
                this.panel.actor.remove_style_class_name('top');

            if(!this.panel.actor.has_style_class_name('bottom'))
                this.panel.actor.add_style_class_name('bottom');
        }

        Main.layoutManager._updateHotCorners();
        Main.layoutManager._updatePanelBarrier(this);
    },

    _removeTopLimit: function() {
        if (this._topLimit) {
            Main.layoutManager.removeChrome(this._topLimit);
            this._topLimit = null;
        }
    },

    _setActivitiesButtonVisible: function(isVisible) {
        if(this.panel.statusArea.activities)
            isVisible ? this.panel.statusArea.activities.actor.show() :
                this.panel.statusArea.activities.actor.hide();
    },

    _setClockLocation: function(loc) {
        if(!this.panel.statusArea.dateMenu)
            return;

        let dateMenuContainer = this.panel.statusArea.dateMenu.container;
        let parent = dateMenuContainer.get_parent();
        let destination;
        let refSibling = null;

        if (!parent) {
            return;
        }

        if (loc.indexOf('BUTTONS') == 0) {
            destination = this.panel._centerBox;
        } else if (loc.indexOf('STATUS') == 0) {
            refSibling = this.panel.statusArea.aggregateMenu ? this.panel.statusArea.aggregateMenu.container : null;
            destination = this.panel._rightBox;
        }

        if (parent != destination) {
            parent.remove_actor(dateMenuContainer);
            destination.add_actor(dateMenuContainer);
        }

        destination['set_child_' + (loc.indexOf('RIGHT') > 0 ? 'above' : 'below') + '_sibling'](dateMenuContainer, refSibling);
    },

    _displayShowDesktopButton: function (isVisible) {
        if(isVisible) {
            if(this._showDesktopButton)
                return;

            this._showDesktopButton = new PanelMenu.Button(null, "Desktop Button");
            let hbox = new St.BoxLayout({ style_class: 'panel-status-menu-box' });
            let icon = new St.Icon({ gicon: Gio.icon_new_for_string(SHOW_DESKTOP_ICON),
                                     style_class: 'system-status-icon' });

            hbox.add_child(icon);
            this._showDesktopButton.actor.add_child(hbox);

            this._showDesktopButton.actor.connect('button-press-event', Lang.bind(this, this._onShowDesktopButtonPress));

            this.panel.addToStatusArea('desktop-button', this._showDesktopButton, this.panel._rightBox.get_n_children());
        } else {
            if(!this._showDesktopButton)
                return;

            this._showDesktopButton.destroy();
            this._showDesktopButton = null;
        }
    },

    _onShowDesktopButtonPress: function() {
        if(this._focusAppChangeId){
            tracker.disconnect(this._focusAppChangeId);
            this._focusAppChangeId = null;
        }

        if(this._restoreWindowList && this._restoreWindowList.length) {
            let current_workspace = Utils.DisplayWrapper.getWorkspaceManager().get_active_workspace();
            let windows = current_workspace.list_windows();
            this._restoreWindowList.forEach(function(w) {
                if(windows.indexOf(w) > -1)
                    Main.activateWindow(w);
            });
            this._restoreWindowList = null;

            Main.overview.hide();
        } else {
            let current_workspace = Utils.DisplayWrapper.getWorkspaceManager().get_active_workspace();
            let windows = current_workspace.list_windows().filter(function (w) {
                return w.showing_on_its_workspace() && !w.skip_taskbar;
            });
            windows = global.display.sort_windows_by_stacking(windows);

            windows.forEach(function(w) {
                w.minimize();
            });
            
            this._restoreWindowList = windows;

            Mainloop.timeout_add(0, Lang.bind(this, function () {
                this._focusAppChangeId = tracker.connect('notify::focus-app', Lang.bind(this, function () {
                    this._restoreWindowList = null;
                }));
            }));

            Main.overview.hide();
        }
    },
});
